@@ -13,9 +13,12 @@
 #'
 #' @example incl/autoStopCluster.R
 #'
-#' @seealso
+#' @details
 #' The cluster is stopped using
 #' \code{\link[parallel:makeCluster]{stopCluster}(cl)}.
+#' An alternative to explicitly call this function on an existing
+#' `cluster` object, is to create the `cluster` object using
+#' `makeClusterPSOCK()` with argument `autoStop = TRUE`.
 #'
 #' @importFrom parallel stopCluster
 #' @importFrom utils capture.output
@@ -31,10 +34,10 @@ autoStopCluster <- function(cl, debug = FALSE) {
 
   if (debug) {
     reg.finalizer(env, function(e) {
-      message("Finalizing cluster ...")
-      message(capture.output(print(e$cluster)))
+      mdebug_push("Finalizing cluster ...")
+      on.exit(mdebug_pop("Finalizing cluster ... done"))
+      mprint(e$cluster)
       try(stopCluster(e$cluster), silent = FALSE)
-      message("Finalizing cluster ... done")
     })
   } else {
     reg.finalizer(env, function(e) {
