@@ -4,7 +4,7 @@
 .onLoad <- function(libname, pkgname) {
   debug <- isTRUE(as.logical(getEnvVar2("R_PARALLELLY_DEBUG", "FALSE")))
   if (debug) options(parallelly.debug = TRUE)
-  debug <- getOption2("parallelly.debug", debug)
+  debug <- getOption("parallelly.debug", debug)
 
   ## Automatically play nice when 'R CMD check' runs?
   if (isTRUE(as.logical(getEnvVar2("R_PARALLELLY_R_CMD_CHECK_NICE", "TRUE"))) && inRCmdCheck()) {
@@ -60,6 +60,9 @@
   }
 
   registerClusterTypes()
+
+  ## R CMD build
+  register_vignette_engine_during_build_only(pkgname)
 }
 
 
